/********************************************************************/
/* COPYRIGHTS:                                                      */
/*                                                                  */
/*   This module contains code made available by IBM Corporation on */
/* an AS IS basis.  Any one receiving the module is  considered  to */
/* be  licensed under IBM copyrights to use the IBM-provided source */
/* code in any way he or  she  deems  fit,  including  copying  it, */
/* compiling  it,  modifying  it,  and  redistributing  it, with or */
/* without modifications.  No license  under  any  IBM  patents  or */
/* patent  applications  is  to  be  implied  from  this  copyright */
/* license.                                                         */
/*                                                                  */
/*   A user of the module should understand that IBM cannot provide */
/* technical support for the module and will not be responsible for */
/* any consequences of use of the program.                          */
/*                                                                  */
/*   Any notice,  including this one, are not to  be  removed  from */
/* the module with out the prior written consent of IBM.            */
/*                                                                  */
// @4  Add DOS error level to match rtopcb and rfrompcb 
// @5  Update comments to specify fully qualified request name.
// @6  Resolve name function
/********************************************************************/


//cwbx.h and cwbx_i.c are in the Express Toolkit include directory.
#include "cwbx.h"		//Express Toolkit automation object MIDL generated header
#include "windows.h"
#include "stdio.h"
#include "iostream.h"
#include "atlbase.h"	
#include "atlconv.h"

//generates IDispatch interface stuff but we'll use vTable
//#import "C:\Program Files\IBM\Client Access\Shared\cwbx.dll"	

//Forward declares
HRESULT PrintErrors(HRESULT FailedHr);	
void	displayUsage();
void	enumDataError(cwbdtDataErrorEnum e, BSTR *);
HRESULT PrintTransferResults(CComPtr<IcwbxDatabaseTransferResults> pIResults);
HRESULT PrintCAErrors(CComPtr<IcwbxErrors> pIcwbxErrors);
BOOLEAN GetCAVer(long& lVer, long& lRel, long& lMod,
				 char * szSP,	 long lenSP);
BOOLEAN resolveName(LPCTSTR szInFile, BSTR * bstrRequest);	//@6
VOID printLastError(char * title);		//@6

///////////////////////////////////////////////////////////////////////////////
//main
///////////////////////////////////////////////////////////////////////////////
void main(int argc, char* argv[])
{
	USES_CONVERSION;	//include the ATL string conversion macros

	BSTR bstrRequest	= NULL; 
	BSTR bstrUser		= NULL; 
	BSTR bstrPwd		= NULL;	
	int  errLevel		= 0;
	
	///////////////////////////////
	//Parse the command line args
	if ( argc != 4 )
	{ 
		displayUsage(); 
		exit(4);
	}
    else
    {
		//@6bstrRequest		= A2BSTR( argv[1] );	//Macro calls SysAllocString 
		if (resolveName(A2T(argv[1]), &bstrRequest) == FALSE)	//@6A resolve unqualifed files with path.
			exit(4);
		bstrUser		= A2BSTR( argv[2] );
		bstrPwd			= A2BSTR( argv[3] );
    }		

	/////////////////
	//Initialize COM
	HRESULT hr = CoInitialize(NULL);
	if (FAILED(hr)) 
	{
		cout << "CoInitialize Failed: " << hr << endl;
		exit(4);
	}

	///////////////////////////////
	//Verify CA Version
	char    szSP[8];
	long	lVer=0, lRel=0, lMod=0;
	if (GetCAVer(lVer, lRel, lMod, szSP, sizeof(szSP)))
	{
		cout << "Client Access Express " << endl;
		cout << "   Version " << lVer 
			 << "  Release " << lRel 
			 << "  Modification " << lMod
			 << "  Service Pack " << szSP << endl << endl;

		//SP2 or later is required or non-admins will get an license error.
		if (lVer < 5				||				//Not V5
			 (lVer == 5 && lRel == 1 &&				//R510 SP 1 or none
					(stricmp(szSP, "SI01037") == 0 || strlen(szSP) == 0)	//TODO: How is none returned?
			 )
			)
		{
			cout << "NOTE:  SI01907 or later may be required when run from a service. " << endl;	
		}
	}


	///////////////////////////////////////////////////////////////////////////////////
	//Create DatabaseTransfer high level object and get pointer to the  xfer interface
	IcwbxDatabaseTransfer * pIcwbxDT;
	hr = CoCreateInstance(CLSID_DatabaseTransfer, NULL, CLSCTX_ALL, 
							IID_IcwbxDatabaseTransfer,
							reinterpret_cast<void **>(&pIcwbxDT));
	if (FAILED(hr)) 
	{
		errLevel = 4;
		cout << "Unable to create cwbx DatabasTransfer object.  rc " << hr << endl;
		cout << "Verify that Client Access Express R510 or later is installed." << endl;	
	}
	else		
	{
		//We have an object.  Run the request....
		DWORD StartTime = GetTickCount();
		hr = pIcwbxDT->put_UserID(bstrUser);		//Set the userid property
		hr = pIcwbxDT->put_Password(bstrPwd);		//Set the password property
		hr = pIcwbxDT->Transfer(bstrRequest);		//Run the request
		if (FAILED(hr)) 
		{
			errLevel = 4;
			cout << "Transfer request " << argv[1] << " FAILED." << endl << endl;
					
			PrintErrors(hr);		//Get the automation error

			//Get the CA Errors 
			IcwbxErrors * pIcwbxErrors;
			hr = pIcwbxDT->get_Errors(&pIcwbxErrors);
			if SUCCEEDED(hr)
			{
				PrintCAErrors(pIcwbxErrors);				
				pIcwbxErrors->Release();
			}
			else
				cout << "Unable to retrieve CAE Error collection.  hr = " << hr << endl;				
		}
		else		//Transfer ran
		{
			if (hr > 0)
				errLevel = 1;		//Set warning.
			DWORD EndTime = GetTickCount();
			cout << "Transfer request " << argv[1] << " SUCCESSFUL."  << endl << endl;
			cout << "   Elasped transfer time: "
				 << (EndTime-StartTime)/(60*60*1000) << " hours "
				 << ((EndTime-StartTime)%(60*60*1000))/60000 << " minutes "
				 << (((EndTime-StartTime)%(60*60*1000))%60000)/1000  << "."
				 << (((EndTime-StartTime)%(60*60*1000))%60000)%1000  << " seconds "
				 << "(" << EndTime-StartTime << "ms)"
				 << endl;
		}		//end of Transfer ran else

		//Get the TransferResult object.  It is set even on failure. 
		IcwbxDatabaseTransferResults * pIResults = NULL;
		HRESULT resultHr = pIcwbxDT->get_TransferResults( &pIResults );
		if (SUCCEEDED(hr))
		{
			PrintTransferResults(pIResults);
			pIResults->Release();
		}				
		
		pIcwbxDT->Release();
	}	//end of CoCreateInstance else

	//Free all Bstr's
	SysFreeString(bstrUser);
	SysFreeString(bstrPwd);
	SysFreeString(bstrRequest);

	CoUninitialize();	
	
	exit( errLevel );		//Use exit to set ERRORLEVEL 
}

///////////////////////////////////////////////////////////////////////////////
//displayUsage()
///////////////////////////////////////////////////////////////////////////////
void displayUsage()
{
	printf("IBM iSeries Support Center sample for unattended batch data transfers.\n");
	printf("   This sample program wraps the R510 CAE data transfer \n");
	printf("   automation objects.  The sample accepts a user ID and \n");
	printf("   password in addition to the transfer request file. \n");
	printf("   Because the prompt mode is set to \"never\" it can  \n");
	printf("   be used in unattended environments such as with CAE  \n");
	printf("   incoming remote command service (IRC) \n");
	printf("   or the Windows scheduler service.\n");
	printf("   \n");
	printf("   This sample is provided on an AS IS basis.\n");
	printf("   A user of this sample should understand that IBM cannot provide\n");
	printf("   technical support for the module and will not be responsible for\n");
	printf("   any consequences of use of the program.\n");
	printf("   Comments concerning this sample can be sent to ddilling@us.ibm.com.\n");
	printf("\n\n");
	printf("To run an unattended transfer w/o prompting of user/pwd:\n\n");		
	
	printf("   RXFERPCB  request userID password\n\n");

	printf("   request  - Fully qualified file name of any Client Access\n");
	printf("              upload or download request of type \n");
	printf("              .DTF, .DTT, .TTO, or .TFR. \n"); 
	printf("   userID   - A valid iSeries user profile for the system \n");
	printf("              specified in the request.\n");
	printf("   password - A valid password for the specified user profile.\n\n");
}

///////////////////////////////////////////////////////////////////////////////
//PrintErrors
//Print the automation error information
///////////////////////////////////////////////////////////////////////////////
HRESULT PrintErrors(HRESULT FailedHr)	
//HRESULT PrintErrors(HRESULT FailedHr,  IUnknown *  pFailedInterface )
{
	USES_CONVERSION;	//include the ATL string conversion macros

	CComPtr<IErrorInfo> pErrorInfo;			//IErrorInfo *          pErrorInfo;
	CComPtr<IErrorInfo> pErrorInfoRec;		//IErrorInfo *          pErrorInfoRec;	
	
	HRESULT               hr;

	//Should check if the failing interface supports Error Info objects
	//however this is a required interface for automation object.
	//CComPtr<ISupportErrorInfo> pSptErrInfo;
	//hr = pFailedInterface->QueryInterface(IID_ISupportErrorInfo, (void **) &pSptErrInfo);
	//hr = pSptErrInfo->InterfaceSupportsErrorInfo(__uuidof(pFailedInterface));
	//if (FAILED(hr))
	//{
	//	printf("Interface does not support Error Info");
	//	return FailedHr;
	//}

	cout << "     hr:          " << FailedHr << endl;	//print out the failed hr
	::GetErrorInfo(0,&pErrorInfo);						// Get the current error object. 
	if (!pErrorInfo) return (FailedHr);					//Return if no error object exists.

	//We got the pointer so print out the description and source
	BSTR bstrDescriptionOfError = NULL;
	BSTR bstrSourceOfError = NULL;
	hr = pErrorInfo->GetDescription(&bstrDescriptionOfError); 
	hr = pErrorInfo->GetSource(&bstrSourceOfError);

	cout << "     Description: " << W2A(bstrDescriptionOfError) << endl;
	cout << "     Source:      " << W2A(bstrSourceOfError) << endl;

	// Free the resources
	SysFreeString(bstrDescriptionOfError);
	SysFreeString(bstrSourceOfError);


	// Smart pointer will release the error object

	return S_OK;
}

///////////////////////////////////////////////////////////////////////////////
//enumDataError - Return a BSTR describing the Data Error enum value
//Note:  Caller must free the BSTR
///////////////////////////////////////////////////////////////////////////////
void enumDataError(cwbdtDataErrorEnum e, BSTR * bstrDataError)
{

/*	enum cwbdtDataErrorEnum
    {	
	cwbdtWarningRounding	= 2,
	cwbdtWarningMissingData	= 3,
	cwbdtWarningUntranslatableField	= 4,
	cwbdtErrorTruncation	= 5,
	cwbdtErrorRounding	= 6,
	cwbdtErrorMissingData	= 7,
	cwbdtErrorUntranslatableField	= 8
    }	cwbdtDataErrorEnum;
*/
	switch (e)
	{
	case cwbdtWarningTruncation:
		*bstrDataError = T2BSTR(_T("cwbdtWarningTruncation = 1"));
		break;
	case cwbdtWarningRounding:
		*bstrDataError = T2BSTR(_T("cwbdtWarningRounding	= 2"));
		break;
	case cwbdtWarningMissingData:
		*bstrDataError = T2BSTR(_T("cwbdtWarningMissingData	= 3"));
		break;
	case cwbdtWarningUntranslatableField:
		*bstrDataError = T2BSTR(_T("cwbdtWarningUntranslatableField	= 4"));
		break;
	case cwbdtErrorTruncation:
		*bstrDataError = T2BSTR(_T("cwbdtErrorTruncation	= 5"));
		break;
	case cwbdtErrorRounding:
		*bstrDataError = T2BSTR(_T("cwbdtErrorRounding	= 6"));
		break;
	case cwbdtErrorMissingData:
		*bstrDataError = T2BSTR(_T("cwbdtErrorMissingData	= 7"));
		break;
	case cwbdtErrorUntranslatableField:
		*bstrDataError = T2BSTR(_T("cwbdtErrorUntranslatableField	= 8"));
		break;
	default:
		*bstrDataError = T2BSTR(_T("Unknown error"));
	}

	return;
}

///////////////////////////////////////////////////////////////////////////////
//PrintCAErrors
//Print each CA error in the errors collection to stdout
///////////////////////////////////////////////////////////////////////////////
HRESULT PrintCAErrors(CComPtr<IcwbxErrors> pIcwbxErrors)
{
	USES_CONVERSION;	//include the ATL string conversion macros
	HRESULT hr;

	//List the Client Access Error collection
	IcwbxError * pIcwbxError = NULL;
	long numCwbxErrs = 0;
	hr = pIcwbxErrors->get_Count(&numCwbxErrs);
	if (numCwbxErrs > 0)
	{
		cout << endl << numCwbxErrs << " Client Access Error Messages:" << endl;
		for (int i = 1; i <= numCwbxErrs; i++)
		{
			//Get the cwbx error object
			hr = pIcwbxErrors->get_Item(i, &pIcwbxError);			
			if (SUCCEEDED(hr))
			{
				//Get the error type
				BSTR bstrMsg = NULL;
				pIcwbxError->get_Text(&bstrMsg);
				cout << "   " << W2A(bstrMsg) << endl;
				SysFreeString(bstrMsg);
				
				//TODO:  Pull the host msg info with a program call
				cout << endl;
				pIcwbxError->Release();	
			}
			else
			{
				cout << "   Unable to retrieve cwbxError object: " << hr << endl;
				break;
			}
		}		
	}

	//smart pointer will release IcwbxErrors
	return hr;
}

///////////////////////////////////////////////////////////////////////////////
//PrintTransferResults
//Print out all the cwbxDatabaseTranferResult information to stdout
///////////////////////////////////////////////////////////////////////////////
HRESULT PrintTransferResults(CComPtr<IcwbxDatabaseTransferResults> pIResults)
{
	USES_CONVERSION;
	HRESULT hr;
			
	//Completion Time
	DATE		CompletionTime = 0;
	SYSTEMTIME	sysTime;	
	hr = pIResults->get_CompletionTime(&CompletionTime);
	if (SUCCEEDED(hr))
	{
		VariantTimeToSystemTime( CompletionTime, &sysTime);              
		cout << "   CompletionTime          = " 
			 << sysTime.wYear << "-" << sysTime.wMonth  << "-" << sysTime.wDay << " " 
			<< sysTime.wHour << ":" << sysTime.wMinute << ":" << sysTime.wSecond << "." << sysTime.wMilliseconds  
			<< endl;
	}
	else
		cout << "Unable to retrieve completion time.  hr = " << hr << endl;

	//Rows xfered
	long rows = 0;
	if (SUCCEEDED( pIResults->get_RowsTransferred(&rows)))		
		cout << "   Rows Transferred        = " << rows << endl;
	else
		cout << "Unable to retrieve RowsTransferred.  hr = " << hr << endl;

	//HTML Tables xfered if the request is a download to html format
	long cols = 0;
	hr = pIResults->get_TablesTransferred(&cols);
	cout << "   HTML Tables Transferred = " << cols << endl;

	//Data error/warnings
	VARIANT_BOOL b = VARIANT_FALSE;
	hr = pIResults->get_DataWarningsOccurred(&b);
	cout << "   Data Warnings Occurred  = " << (b == VARIANT_TRUE ? "TRUE": "FALSE")  << endl;
	hr = pIResults->get_DataErrorsOccurred(&b);
	cout << "   Data Errors Occurred    = " << (b == VARIANT_TRUE ? "TRUE": "FALSE")  << endl;

	//List the data error/warning messages.  This is specific to the data as opposed	
	//to cwbError which is for more general errors like the request failing to run.
	IcwbxDatabaseDataErrors * pIDataErrors = NULL;
	hr = pIResults->get_DataErrors(&pIDataErrors);
	if (SUCCEEDED(hr))
	{
		IcwbxDatabaseDataError * pIDataError = NULL;
		long NumDataErrs = 0;
		pIDataErrors->get_Count(&NumDataErrs);
		if (NumDataErrs > 0)
		{
			cout << "   Data Error/Warning Messages:" << endl;
			for (int i = 1; i <= NumDataErrs; i++)
			{
				//Get the data error object
				hr = pIDataErrors->get_Item(i, &pIDataError);
				if (SUCCEEDED(hr))
				{
					//Get the error type
					cwbdtDataErrorEnum  DataErrType;
					pIDataError->get_ErrorType(&DataErrType);
					BSTR bstrDataErrType = NULL;
					enumDataError(DataErrType, &bstrDataErrType);
					cout << "      Error Type:    " << W2A(bstrDataErrType) << endl;
					SysFreeString(bstrDataErrType);
					
					//The row
					long row = 0;
					pIDataError->get_Row(&row);			
					cout << "      Error Row:     " << row << endl;

					//the column
					long col = 0;
					pIDataError->get_Column(&col);			
					cout << "      Error Column:  " << col << endl;
					cout << endl;
					pIDataError->Release();
				}
				else
				{
					cout << "Unable to get DatabaseDataError object.  hr = " << hr << endl;
					break;
				}
			}	//for each data error...
		}	//end if (NumDataErrs > 0)
		pIDataErrors->Release();		//Release the collection
	}	//end if for hr = pIResults->get_DataErrors(&pIDataErrors);

	return hr;
}

///////////////////////////////////////////////////////////////////////////////
//VerifyCAVer
//TODO
///////////////////////////////////////////////////////////////////////////////
BOOLEAN GetCAVer(long& lVer, long& lRel, long& lMod,
				 char * szSP,	 long lenSP)
{
	//Since we're using automation anyway we'll use the Client Info object
	
	USES_CONVERSION;
	BOOLEAN  bInError = FALSE;

	//Validate input args
	if (lenSP < 8) return FALSE;	//SI01907  7 + null

	//Initialize the input vars
	lVer	= 0;
	lRel	= 0;
	lMod	= 0;
	*szSP	= '\0';

	//Get the Client Info object
	IcwbxClientInfo * pIcwbxClientInfo;
	HRESULT hr = CoCreateInstance(CLSID_ClientInfo, NULL, CLSCTX_ALL,
							IID_IcwbxClientInfo,
							reinterpret_cast<void **>(&pIcwbxClientInfo));
	if (SUCCEEDED(hr)) 
	{
		
		if ( FAILED( pIcwbxClientInfo->get_ClientVersion(     &lVer)	  ) )
			bInError = TRUE;
		if ( FAILED( pIcwbxClientInfo->get_ClientRelease(     &lRel)	  ) ) 
			bInError = TRUE;
		if ( FAILED( pIcwbxClientInfo->get_ClientModificationLevel(&lMod) ) ) 
			bInError = TRUE;

		BSTR  bstrSP = NULL;
		if ( FAILED( pIcwbxClientInfo->get_ClientFixLevel(&bstrSP)  ) ) 
			bInError = TRUE;
		else
		{
			strcpy(szSP, OLE2A(bstrSP));
			::SysFreeString(bstrSP);
		}
		pIcwbxClientInfo->Release();
	}
	else	//Failed to get interface
		bInError = TRUE;

	if (bInError) 
		return FALSE;
	else
		return TRUE;	
}	//end GetCAVer

/////
//@6  The automation objects return "Invalid transfer request" if the file
//    is not found.  Added this function to verify file exists and resolve
//    the path if the user did not fully qualify the name.
//
//    return FALSE for error such as file not found.
//    return TRUE and allocated BSTR with the qualified file name othewise.
////
BOOLEAN resolveName(LPCTSTR  szInFile, 
					BSTR * bstrRequest)
					
{
	LPTSTR pszFilePart = NULL;
	TCHAR szOutFile[MAX_PATH + 1] = _T("\0");	
	unsigned int nOutFileCharCount = MAX_PATH;	
	*bstrRequest = NULL;					//Initialize to null

	//Searchpath returned a invalid parameter error for a fully 
	//qualifed path that does not exist.  That's confusing too so use this 
	//if we have a fully qualified path.
	if (_tcsclen(szInFile) > 2 &&
		szInFile[1] == _T(':') &&
		szInFile[2] == _T('\\') )
	{
		
		HANDLE hFile = CreateFile(szInFile,		// open MYFILE.TXT 
                GENERIC_READ,					// open for reading 
                FILE_SHARE_READ,				// share for reading 
                NULL,							// no security 
                OPEN_EXISTING,					// existing file only 
                FILE_ATTRIBUTE_NORMAL,			// normal file 
                NULL); 
		if (hFile == INVALID_HANDLE_VALUE)
		{
			printLastError("\nOpen of transfer request");			
			return FALSE;
		}
		else
		{
			CloseHandle(hFile);
			* bstrRequest = T2BSTR( szInFile );	//sysalloc
			return TRUE;
		}		
	}

	//We must have an unqualified file name.  Try to find it...
	nOutFileCharCount = SearchPath(
					  NULL,					// search path - NULL is default path
					  szInFile,				// file name
					  NULL,					// file extension - NULL requires user to enter it.
					  nOutFileCharCount,	// size of buffer in TCHARs
					  szOutFile,			// found file name buffer
					  &pszFilePart			// file component
					);
		
	// returns count if successful
	if ( nOutFileCharCount > 0)					
	{
		//SearchPath appears to return a null terminated string but the doc doesn't
		//explicitly state that so I'm not trusting it.
		*(szOutFile + nOutFileCharCount) = _T('\0');		//Null terminate it.
		* bstrRequest = T2BSTR( szOutFile );				//sysalloc
		return TRUE;
	}
	else
	{
		printLastError("\nOpen of transfer request");
		return FALSE;
	}

}	//resolveName

//@6A
VOID printLastError(char * title)
{
	int		iLastErr = 0;
	int		fmtrc;
	PVOID	lpMsgBuf;

	iLastErr = GetLastError();
	if (iLastErr == 0)
		printf("%s failed with an Unknown error",
				title);
	else
	{
		fmtrc = FormatMessage( 
					 FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
					 NULL,
					 iLastErr,
					 MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
					 (LPTSTR) &lpMsgBuf,
					 0,
					 NULL );
		if (fmtrc > 0)
		{
			// Display the string.
			printf("%s failed with return code %d: ""%s""\n\r",
						title, iLastErr, lpMsgBuf);
			// Free the buffer.
			LocalFree( (HLOCAL) lpMsgBuf );
		}
		else
			printf("%s failed with return code %d: Unknown error\n\r",
						title, iLastErr);
	}
return;
}